/*
   AtpImage.cpp - implementation of the ATP image format

   Copyright (C) 2003 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "AtpImage.h"
#include "Indent.h"
#include "SIOTracer.h"
#include "AtariDebug.h"

AtpImage::AtpImage(size_t numberOfTracks)
	: fNumberOfTracks(numberOfTracks)
{
	AllocData();
}

AtpImage::~AtpImage()
{
	FreeData();
}

void AtpImage::AllocData()
{
	if (fNumberOfTracks) {
		fTracks = new AtpTrack[fNumberOfTracks];
		for (size_t i=0;i<fNumberOfTracks;i++) {
			fTracks[i].SetTrackNumber(i);
		}
	} else {
		fTracks = 0;
	}
}

void AtpImage::FreeData()
{
	if (fTracks) {
		delete[] fTracks;
	}
	fTracks = 0;
	fNumberOfTracks = 0;
}

bool AtpImage::SetDensity(Atari1050Model::EDiskDensity dens, size_t trackno)
{
	if (trackno < fNumberOfTracks) {
		fTracks[trackno].SetDensity(dens);
		SetChanged(true);
		return true;
	} else {
		return false;
	}
}

bool AtpImage::SetDensity(Atari1050Model::EDiskDensity dens)
{
	for (size_t i=0;i<fNumberOfTracks;i++) {
		fTracks[i].SetDensity(dens);
	}
	return true;
}

Atari1050Model::EDiskDensity AtpImage::GetDensity(size_t trackno) const
{
	if (trackno < fNumberOfTracks) {
		return fTracks[trackno].GetDensity();
	} else {
		return Atari1050Model::eDensityFM;
	}
}

void AtpImage::SetNumberOfTracks(size_t tracks)
{
	FreeData();
	fNumberOfTracks = tracks;
	AllocData();
	SetChanged(true);
}

bool AtpImage::AddSector(size_t trackno, const RCPtr<AtpSector>& sector)
{
	if (trackno < fNumberOfTracks) {
		fTracks[trackno].AddSector(sector);
		return true;
	} else {
		return false;
	}
	SetChanged(true);
}

bool AtpImage::GetSector(size_t trackno,
		size_t sectorID,
		RCPtr<AtpSector>& sector,
		size_t current_time) const
{
	if (trackno < fNumberOfTracks) {
		return fTracks[trackno].GetSector(sectorID, sector, current_time);
	} else {
		sector = RCPtr<AtpSector>();
		return false;
	}
}

void AtpImage::Dump(std::ostream& os, size_t indentlevel)
{
	using std::endl;
	os << Indent(indentlevel)
	   << "begin AtpImage {"
	   << endl
	;
	os << Indent(indentlevel+1)
	   << "Tracks: " << fNumberOfTracks
	   << endl
	;
	os << Indent(indentlevel+1)
	   << "write protected: "
	;
        if (IsWriteProtected()) {
		os << "yes";
	} else {
		os << "no";
	}
	os << endl;
	for (size_t i=0;i<fNumberOfTracks;i++) {
		fTracks[i].Dump(os, indentlevel+1);
	}
	os << Indent(indentlevel)
	   << "} // end AtpImage"
	   << endl
	;
}

RCPtr<ChunkWriter> AtpImage::BuildHeaderChunk() const
{
	RCPtr<ChunkWriter> headerChunk(new ChunkWriter("INFO"));

	headerChunk->AppendDword(fNumberOfTracks);
	
	if (IsWriteProtected()) {
		headerChunk->AppendDword(1);
	} else {
		headerChunk->AppendDword(0);
	}

	headerChunk->CloseChunk();
	return headerChunk;
}

bool AtpImage::InitFromHeaderChunk(RCPtr<ChunkReader> chunk)
{
	if (!chunk) {
		return false;
	}

	if (strcmp(chunk->GetChunkName(),"INFO") == 0) {
		size_t tmp;
		if (!chunk->ReadDword(tmp)) {
			return false;
		}
		SetNumberOfTracks(tmp);
		if (!chunk->ReadDword(tmp)) {
			return false;
		}
		if (tmp==0) {
			SetWriteProtect(false);
		} else {
			SetWriteProtect(true);
		}

		return true;

	} else {
		return false;
	}
}

bool AtpImage::WriteImageToFile(const char* filename) const
{
	RCPtr<FileIO> fileio;

#ifdef USE_ZLIB 
	fileio = new GZFileIO();
#else   
	fileio = new StdFileIO();
#endif  

	int len = strlen(filename);

	if (len > 3 && strcasecmp(filename+len-3,".gz") == 0) {
#ifdef USE_ZLIB
		fileio = new GZFileIO();
#else
		ERROR("cannot write comressed file - zlib support disabled at compiletime!");
		return false;
#endif
	} else {
		fileio = new StdFileIO();
	}

	if (!fileio->OpenWrite(filename)) {
		ERROR("cannot create %s", filename);
		return false;
	}

	RCPtr<ChunkWriter> fileChunk(new ChunkWriter("FORM"));

	{
		RCPtr<ChunkWriter> atpChunk(new ChunkWriter("ATP1"));

		atpChunk->AppendChunk(BuildHeaderChunk());

		for (size_t i=0; i<fNumberOfTracks;i++) {
			atpChunk->AppendChunk(fTracks[i].BuildTrackChunk());
		}

		size_t checksum = atpChunk->CalculateCRC32();

		atpChunk->AppendDword(checksum);

		atpChunk->CloseChunk();

		fileChunk->AppendChunk(atpChunk);
	}

	fileChunk->CloseChunk();

	if (!fileChunk->WriteToFile(fileio)) {
		ERROR("writing atp image failed!");
		fileio->Close();
		return false;
	}
	fileio->Close();
	SetChanged(false);
	return true;
}

bool AtpImage::ReadImageFromFile(const char* filename, bool beQuiet)
{
	RCPtr<FileIO> fileio;

	FreeData();

#ifdef USE_ZLIB 
	fileio = new GZFileIO();
#else   
	fileio = new StdFileIO();
#endif  

	if (!fileio->OpenRead(filename)) {
		if (!beQuiet) {
			ERROR("cannot open \"%s\" for reading",filename);
		}
		return false;
	}

	RCPtr<ChunkReader> fileChunk(ChunkReader::OpenChunkFile(fileio));
	RCPtr<ChunkReader> formChunk;
	RCPtr<ChunkReader> atpChunk;

	if (!fileChunk) {
		goto error;
	}
	
	do {
		formChunk = fileChunk->OpenChunk();
	} while (formChunk && strcmp(formChunk->GetChunkName(),"FORM"));

	if (!formChunk) {
		if (!beQuiet) {
			ERROR("cannot find FORM chunk in file");
		}
		goto error;
	}
	
	do {
		atpChunk = formChunk->OpenChunk();
	} while (atpChunk && strcmp(atpChunk->GetChunkName(),"ATP1"));

	if (!atpChunk) {
		if (!beQuiet) {
			ERROR("cannot find ATP1 chunk in file");
		}
		goto error;
	}

	{
		RCPtr<ChunkReader> headerChunk;
		
		size_t chksum_start;
		do {
			chksum_start = atpChunk->GetCurrentPosition();
			headerChunk = atpChunk->OpenChunk();
		} while (headerChunk && strcmp(headerChunk->GetChunkName(),"INFO"));

		if (!headerChunk) {
			if (!beQuiet) {
				ERROR("cannot find INFO chunk in file");
			}
			goto error;
		}

		if (!InitFromHeaderChunk(headerChunk)) {
			if (!beQuiet) {
				DPRINTF("initialization of AtpImage from INFO chunk failed");
			}
			goto error;
		}

		for (size_t i=0; i<fNumberOfTracks; i++) {
			RCPtr<ChunkReader> trackChunk;

			do {
				trackChunk = atpChunk->OpenChunk();
			} while (trackChunk && strcmp(trackChunk->GetChunkName(),"TRAK"));

			if (!trackChunk) {
				if (!beQuiet) {
					ERROR("cannot find TRAK (%d) chunk in file", i);
				}
				goto error;
			}

			if (!fTracks[i].InitFromTRAKChunk(trackChunk, beQuiet)) {
				if (!beQuiet) {
					DPRINTF("initialization of track %d from TRAK chunk failed",i);
				}
				goto error;
			}
		}

		size_t chksum_end = atpChunk->GetCurrentPosition();

		size_t file_chksum;
		size_t calculated_chksum;

		if (!atpChunk->ReadDword(file_chksum)) {
			goto error;
		}

		if (!atpChunk->CalculateCRC32(calculated_chksum, chksum_start, chksum_end)) {
			if (!beQuiet) {
				DPRINTF("calculating the checksum failed");
			}
			goto error;
		}

		if (file_chksum != calculated_chksum) {
			if (!beQuiet) {
				ERROR("checksum error");
			}
			goto error;
		}
	}

	fileio->Close();
	SetChanged(false);
	return true;
error:
	fileio->Close();
	return false;

}

bool AtpImage::InitBlankSD()
{
	FreeData();
	fNumberOfTracks = 40;
	AllocData();
	unsigned char buf[128];
	memset(buf,0,128);
	for (size_t track=0;track<40;track++) {
		for (size_t sector=1;sector<=18;sector++) {
			size_t position = Atari1050Model::CalculatePositionOfSDSector(
				track, sector);
			AddSector(track, new AtpSector(sector, 128, buf, position, Atari1050Model::eSDSectorTimeLength));
		}
	}
	SetDensity(Atari1050Model::eDensityFM);
	SetChanged(true);
	return true;
}

bool AtpImage::InitBlankED()
{
	FreeData();
	fNumberOfTracks = 40;
	AllocData();
	unsigned char buf[128];
	memset(buf,0,128);
	for (size_t track=0;track<40;track++) {
		for (size_t sector=1;sector<=26;sector++) {
			size_t position = Atari1050Model::CalculatePositionOfEDSector(
				track, sector);
			AddSector(track, new AtpSector(sector, 128, buf, position, Atari1050Model::eEDSectorTimeLength));
		}
	}
	SetDensity(Atari1050Model::eDensityMFM);
	SetChanged(true);
	return true;
}

// currently only returns 90k or 130k (not the real allocated size),
// depending on the format of the first track
size_t AtpImage::GetImageSize() const
{
	if (fNumberOfTracks > 0 && 
	    fTracks[0].GetDensity() == Atari1050Model::eDensityMFM) {
		return 1040*128;
	} else {
		return 720*128;
	}
}

size_t AtpImage::GetNumberOfSectors() const
{
	if (fNumberOfTracks > 0) {
	    if (fTracks[0].GetDensity() == Atari1050Model::eDensityMFM) {
			return 1040;
		} else {
			return 720;
		}
	} else {
		return 0;
	}
}

bool AtpImage::IsAtpImage() const
{
	return true;
}

bool AtpImage::ReadSector(size_t sector, unsigned char* buffer, size_t buffer_length) const
{
	if (sector < 1) {
		return false;
	}

	if (buffer_length != 128) {
		return false;
	}

	RCPtr<AtpSector> sec;

	switch (GetDensity()) {
	case Atari1050Model::eDensityFM:
		if (sector > 720) {
			return false;
		}
		if (!GetSector( (sector - 1) / 18, ( ( sector - 1) % 18 ) + 1, sec)) {
			return false;
		}
		break;
	case Atari1050Model::eDensityMFM:
		if (sector > 1040) {
			return false;
		}
		if (!GetSector( (sector - 1) / 26, ( ( sector - 1) % 26 ) + 1, sec)) {
			return false;
		}
		break;
	default:
		return false;
	}
	if (sec->GetSectorStatus() != 255) {
		return false;
	}
	return sec->GetData(buffer, buffer_length);
}
