/*
   AtpSector.cpp - data of an sector according to the ATP format

   Copyright (C) 2003 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "AtpSector.h"
#include "Indent.h"
#include "Atari1050Model.h"
#include "AtariDebug.h"

AtpSector::AtpSector(
		size_t id,
		size_t data_len,
		const unsigned char* data,
	       	size_t pos,
		size_t time_len,
		unsigned char status)
	: fID(id),
	  fDataLength(data_len),
	  fPosition(pos),
	  fTimeLength(time_len),
	  fSectorStatus(status)
{
	if (fDataLength) {
		fSectorData = new unsigned char[fDataLength];
		memcpy(fSectorData, data, fDataLength);
	} else {
		fSectorData = 0;
	}
}

AtpSector::AtpSector()
	: fID(0),
	  fDataLength(0),
	  fSectorData(0),
	  fPosition(0),
	  fTimeLength(0),
	  fSectorStatus(255)
{
}

AtpSector::~AtpSector()
{
	if (fSectorData) {
		delete[] fSectorData;
	}
	fSectorData = 0;
	fDataLength = 0;
}

bool AtpSector::GetData(unsigned char* dest, size_t len) const
{
	if (len != fDataLength) {
		return false;
	} else {
		memcpy(dest, fSectorData, len);
		return true;
	}
}

bool AtpSector::SetData(const unsigned char* src, size_t len)
{
	if (len != fDataLength) {
		return false;
	} else {
		memcpy(fSectorData, src, len);
		return true;
	}
}

void AtpSector::Dump(std::ostream& os, size_t indentlevel) const
{
	using std::endl;
	os << Indent(indentlevel)
	   << "id: "
	   << fID
	   << " datalen: "
	   << fDataLength
	   << " pos: "
	   << fPosition
	   << " timelen: "
	   << fTimeLength
	   << " status: "
	;
	os.setf(std::ios::hex, std::ios::basefield);
	os.setf(std::ios::showbase);
	os
	   << (size_t)fSectorStatus
	   << endl
	;
	os.setf(std::ios::dec, std::ios::basefield);
}

RCPtr<ChunkWriter> AtpSector::BuildSectorChunk() const
{
	RCPtr<ChunkWriter> chunk(new ChunkWriter("SECT"));

	chunk->AppendDword(fID);
	chunk->AppendDword(fDataLength);
	chunk->AppendByte(fSectorStatus);
	chunk->AppendDword(fPosition);
	chunk->AppendDword(fTimeLength);
	chunk->AppendBlock(fSectorData, fDataLength);

	chunk->CloseChunk();

	return chunk;
}

bool AtpSector::InitFromSectorChunk(RCPtr<ChunkReader> chunk, bool beQuiet)
{
	if (fSectorData) {
		delete[] fSectorData;
		fSectorData = 0;
		fDataLength = 0;
	}

	if (!chunk->ReadDword(fID)) return false;
	if (fID < 1 || fID > 26) {
		if (!beQuiet) {
			DPRINTF("illegal sector ID %u", fID);
		}
		return false;
	}
	if (!chunk->ReadDword(fDataLength)) return false;
	if (fDataLength != 128) {
		if (!beQuiet) {
			DPRINTF("illegal sector data length %u", fDataLength);
		}
		return false;
	}
	if (!chunk->ReadByte(fSectorStatus)) return false;
	if (!chunk->ReadDword(fPosition)) return false;
	if (fPosition >= Atari1050Model::eDiskRotationTime) {
		if (!beQuiet) {
			DPRINTF("illegal sector position %u", fPosition);
		}
		return false;
	}
	if (!chunk->ReadDword(fTimeLength)) return false;
	if (fTimeLength >= Atari1050Model::eDiskRotationTime) {
		if (!beQuiet) {
			DPRINTF("illegal sector time length %u", fTimeLength);
		}
		return false;
	}

	try {
		fSectorData = new unsigned char[fDataLength];
	}
	catch(...) {
		fSectorData = 0;
		if (!beQuiet) {
			DPRINTF("allocating sector data failed");
		}
		return false;
	}
	if (!chunk->ReadBlock(fSectorData, fDataLength)) {
		delete[] fSectorData;
		fSectorData = 0;
		return false;
	}
	return true;
}
