/*
   CursesFrontend.cpp - curses frontend for atariserver

   Copyright (C) 2003 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "CursesFrontend.h"
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>

#include "FileInput.h"
#include "DiskManager.h"
#include "SIOTracer.h"
#include "AtariDebug.h"
#include "FileSelect.h"
#include "MiscUtils.h"
#include "Version.h"

using namespace MiscUtils;

CursesFrontend::CursesFrontend(RCPtr<DiskManager>& manager, bool useColor)
	: fDiskManager(manager),
	  fTraceLevel(0),
	  fDirCache(new DirectoryCache),
	  fCursorStatus(true),
	  fAuxWindowStatus(true),
	  fGotSigWinCh(false)
{
	// init curses

	initscr();
	nonl();
	cbreak();
	noecho();

	ShowCursor(false);

	int tops, toph, drives, driveh, logts, logth;
	int logs, logh, bots, both, inps, inph;
	int auxs, auxh;

	// create windows
	CalculateWindowPositions(
		tops, toph, drives, driveh, logts, logth,
		logs, logh, bots, both, inps, inph, auxs, auxh);

	fTopLineWindow = newwin(toph, fScreenWidth, tops, 0);
        fDriveStatusWindow = newwin(driveh, fScreenWidth, drives, 0);
	fStatusLineWindow = newwin(logth, fScreenWidth, logts, 0);
	fLogWindow = newwin(logh, fScreenWidth, logs, 0);
	fBottomLineWindow = newwin(both, fScreenWidth, bots, 0);
	fInputLineWindow = newwin(inph, fScreenWidth, inps, 0);
	fAuxWindow = newwin(auxh, fScreenWidth, auxs, 0);

	/*
	leaveok(fTopLineWindow, TRUE);
	leaveok(fDriveStatusWindow, TRUE);
	leaveok(fStatusLineWindow, TRUE);
	leaveok(fLogWindow, TRUE);
	leaveok(fBottomLineWindow, TRUE);
	leaveok(fAuxWindow, TRUE);
	*/

        fTopLinePanel = new_panel(fTopLineWindow);
	fDriveStatusPanel = new_panel(fDriveStatusWindow);
	fStatusLinePanel = new_panel(fStatusLineWindow);
	fLogPanel = new_panel(fLogWindow);
	fBottomLinePanel = new_panel(fBottomLineWindow);
	fInputLinePanel = new_panel(fInputLineWindow);
	fAuxPanel = new_panel(fAuxWindow);

	ShowAuxWindow(false);

	// init colors
	if (useColor && has_colors()) {
		start_color();
		init_pair(1, COLOR_WHITE, COLOR_BLUE);
		init_pair(2, COLOR_YELLOW, COLOR_BLUE);
		init_pair(3, COLOR_YELLOW, COLOR_BLACK);
		init_pair(4, COLOR_GREEN, COLOR_BLACK);
		init_pair(5, COLOR_CYAN, COLOR_BLACK);
		init_pair(6, COLOR_RED, COLOR_BLACK);
		init_pair(7, COLOR_YELLOW, COLOR_RED);
		init_pair(8, COLOR_MAGENTA, COLOR_BLACK);

		fLogColorStandard = COLOR_PAIR(0);
		fLogColorHighlight = COLOR_PAIR(3) | A_BOLD;
		fLogColorOK = COLOR_PAIR(4);
		fLogColorDebug = COLOR_PAIR(8);
		fLogColorWarning = COLOR_PAIR(5);
		fLogColorError = COLOR_PAIR(6);

		fDriveColorStandard = COLOR_PAIR(0);
		fDriveColorWritable = COLOR_PAIR(4);
		fDriveColorWriteProtected = COLOR_PAIR(6);
		fDriveColorChanged = COLOR_PAIR(6);

		fStatusColorStandard = COLOR_PAIR(2) | A_BOLD;

		fAuxColorStandard = COLOR_PAIR(0);
		fAuxColorSelected = COLOR_PAIR(7) | A_BOLD;

		wbkgdset(fTopLineWindow, COLOR_PAIR(2) | A_BOLD | ' ');
		wbkgdset(fBottomLineWindow, COLOR_PAIR(1) | ' ');

	} else {
		fLogColorStandard = 0;
		fLogColorOK = 0;
		fLogColorDebug = 0;
		fLogColorWarning = 0;
		fLogColorError = 0;

		fDriveColorStandard = 0;
		fDriveColorWritable = 0;
		fDriveColorWriteProtected = 0;
		fDriveColorChanged = 0;

		fStatusColorStandard = A_REVERSE;

		fAuxColorStandard = 0;
		fAuxColorSelected = A_REVERSE;

		wbkgdset(fTopLineWindow, A_REVERSE | ' ');
		wbkgdset(fBottomLineWindow, A_REVERSE | ' ');
	}

	wbkgdset(fStatusLineWindow, fStatusColorStandard | ' ');
	wbkgdset(fAuxWindow, fAuxColorStandard | ' ');

	scrollok(fLogWindow, TRUE);
	keypad(fInputLineWindow, TRUE);

	InitWindows();
}

CursesFrontend::~CursesFrontend()
{
	del_panel(fTopLinePanel);
	del_panel(fDriveStatusPanel);
	del_panel(fStatusLinePanel);
	del_panel(fLogPanel);
	del_panel(fBottomLinePanel);
	del_panel(fInputLinePanel);
	del_panel(fAuxPanel);

	delwin(fTopLineWindow);
	delwin(fDriveStatusWindow);
	delwin(fStatusLineWindow);
	delwin(fLogWindow);
	delwin(fBottomLineWindow);
	delwin(fInputLineWindow);
	delwin(fAuxWindow);

	endwin();
}

bool CursesFrontend::CalculateWindowPositions(
	int& toplineStart, int& toplineHeight,
	int& drivestatusStart, int& drivestatusHeight,
	int& logtitleStart, int& logtitleHeight,
	int& logStart, int& logHeight,
	int& bottomStart, int& bottomHeight,
	int& inputStart, int& inputHeight,
	int& auxStart, int& auxHeight)
{
	getmaxyx(stdscr, fScreenHeight, fScreenWidth);
	DPRINTF("new width: %d height: %d", fScreenWidth, fScreenHeight);

	if (fScreenHeight < 15) {
		return false;
	}

	toplineStart = 0; toplineHeight = 1;

	drivestatusStart = toplineStart + toplineHeight;
	drivestatusHeight = 10;

	logtitleStart = drivestatusStart + drivestatusHeight;
	logtitleHeight = 1;

	inputHeight = 1;
	inputStart = fScreenHeight - inputHeight;

	bottomHeight = 1;
	bottomStart = inputStart - bottomHeight;

	logStart = logtitleStart + logtitleHeight;
	logHeight = bottomStart - logStart;

	auxStart = drivestatusStart;
	auxHeight = bottomStart - auxStart;

	fDriveYStart = 1;
	fDriveNoX = 0;

	fDriveSectorsX = fDriveNoX + 4;
	fDriveDensityX = fDriveSectorsX + 6;

	fDriveProtectedX = fDriveDensityX + 3;
	fDriveChangedX = fDriveProtectedX + 2;

	fDriveFilenameX = fDriveChangedX + 3;
	if (fDriveFilenameX < fScreenWidth) {
		fDriveFilenameLength = fScreenWidth - fDriveFilenameX;
	} else {
		fDriveFilenameLength = 0;
	}
	return true;
}

void CursesFrontend::UpdateScreen()
{
	update_panels();
	doupdate();
}

void CursesFrontend::HandleResize(bool redrawDriveStatusOnly)
{
	if (!fGotSigWinCh) {
		return;
	}
	fGotSigWinCh = false;

	//endwin();
	//refresh();

	// try to determine new window size
	struct winsize ws;
	int fd;
	if ( (fd = open("/dev/tty", O_RDONLY)) != -1) {
		if (ioctl(fd, TIOCGWINSZ, &ws) != -1) {
			resizeterm(ws.ws_row, ws.ws_col);
		}
		close(fd);
	}

	int tops, toph, drives, driveh, logts, logth;
	int logs, logh, bots, both, inps, inph;
	int auxs, auxh;

	// resize windows
	if ( CalculateWindowPositions(
		tops, toph, drives, driveh, logts, logth,
		logs, logh, bots, both, inps, inph, auxs, auxh) ) {

		wresize(fTopLineWindow, toph, fScreenWidth);
		wresize(fDriveStatusWindow, driveh, fScreenWidth);
		wresize(fStatusLineWindow, logth, fScreenWidth);
		wresize(fLogWindow, logh, fScreenWidth);
		wresize(fBottomLineWindow, both, fScreenWidth);
		wresize(fInputLineWindow, inph, fScreenWidth);
		wresize(fAuxWindow, auxh, fScreenWidth);
	
		replace_panel(fTopLinePanel, fTopLineWindow);
		replace_panel(fDriveStatusPanel, fDriveStatusWindow);
		replace_panel(fStatusLinePanel, fStatusLineWindow);
		replace_panel(fLogPanel, fLogWindow);
		replace_panel(fBottomLinePanel, fBottomLineWindow);
		replace_panel(fInputLinePanel, fInputLineWindow);
		replace_panel(fAuxPanel, fAuxWindow);

		move_panel(fTopLinePanel, tops, 0);
		move_panel(fDriveStatusPanel, drives, 0);
		move_panel(fStatusLinePanel, logts, 0);
		move_panel(fLogPanel, logs, 0);
		move_panel(fBottomLinePanel, bots, 0);
		move_panel(fInputLinePanel, inps, 0);
		move_panel(fAuxPanel, auxs, 0);
	}

	InitDriveStatusWindows();
	if (!redrawDriveStatusOnly) {
		InitTopAndBottomLines();
	}

	bool curs = ShowCursor(!fCursorStatus);
	ShowCursor(curs);

	UpdateScreen();
}

void CursesFrontend::InitWindows()
{
	InitDriveStatusWindows();
	InitTopAndBottomLines();
	bool curs = ShowCursor(!fCursorStatus);
	ShowCursor(curs);
	UpdateScreen();
}

void CursesFrontend::InitDriveStatusWindows()
{
	werase(fDriveStatusWindow);
	werase(fStatusLineWindow);
	werase(fLogWindow);

	fFirstLogLine = true;

	DisplayStatusLine();
	DisplayDriveStatus();
}

void CursesFrontend::InitTopAndBottomLines()
{
	werase(fTopLineWindow);
	werase(fBottomLineWindow);
	werase(fInputLineWindow);
	werase(fAuxWindow);
	InitTopLine();
	ShowStandardHint();
}

bool CursesFrontend::ShowAuxWindow(bool on)
{
	bool oldStat = fAuxWindowStatus;
	if (on) {
		show_panel(fAuxPanel);
		top_panel(fAuxPanel);
	} else {
		hide_panel(fAuxPanel);
	}
	top_panel(fInputLinePanel);
	fAuxWindowStatus = on;
	return oldStat;
}

void CursesFrontend::InitTopLine()
{
	SetTopLineFilename(0, true);
}

void CursesFrontend::SetTopLineString(const char* string)
{
	werase(fTopLineWindow);
	mvwaddstr(fTopLineWindow, 0, 0, "atariserver " VERSION_STRING " -- ");
	unsigned int y,x;
	getyx(fTopLineWindow, y, x);

	if (string && (x < fScreenWidth)) {
		unsigned int len = strlen(string);
		if (x+len > fScreenWidth) {
			len = fScreenWidth - x;
		}
		waddnstr(fTopLineWindow, string, len);
	}
}

void CursesFrontend::SetTopLineFilename(const char* string, bool appendSlash)
{
	werase(fTopLineWindow);
	mvwaddstr(fTopLineWindow, 0, 0, "atariserver " VERSION_STRING " -- ");
	unsigned int y,x;
	getyx(fTopLineWindow, y, x);

	if (x < fScreenWidth) {

		char cwd[PATH_MAX+1];
		if (string == NULL) {
			cwd[0] = 0;
			getcwd(cwd, PATH_MAX);
			cwd[PATH_MAX] = 0;
			string = cwd;
		}

		char* str;

		unsigned int len = strlen(string);
		if (appendSlash && (len > 0) && (string[len-1] != '/') && (len < PATH_MAX)) {
			char* str2 = new char[len+2];
			strcpy(str2, string);
			str2[len++] = '/';
			str2[len] = 0;
			str = ShortenFilename(str2, fScreenWidth - x);
			delete[] str2;
		} else {
			str = ShortenFilename(string, fScreenWidth - x);
		}

		if (str) {
			waddstr(fTopLineWindow, str);
			delete[] str;
		}
	}
}

void CursesFrontend::DisplayStatusLine()
{
	wmove(fStatusLineWindow, 0, 0);
	waddstr(fStatusLineWindow, " cable type: ");

	switch (fDiskManager->GetSioServerMode()) {
	case SIOWrapper::eCommandLine_RI:
		waddstr(fStatusLineWindow, "RING"); break;
	case SIOWrapper::eCommandLine_DSR:
		waddstr(fStatusLineWindow, "DSR"); break;
	case SIOWrapper::eCommandLine_CTS:
		waddstr(fStatusLineWindow, "CTS"); break;
	default:
		waddstr(fStatusLineWindow, "???"); break;
	}

	waddstr(fStatusLineWindow, "  speed: ");
	switch (fDiskManager->GetHighSpeedMode()) {
	case DiskManager::eHighSpeedOff:
		waddstr(fStatusLineWindow, "low");
		break;
	case DiskManager::eHighSpeedOn:
		waddstr(fStatusLineWindow, "high");
		break;
	case DiskManager::eHighSpeedWithPause:
		waddstr(fStatusLineWindow, "high/pause");
		break;
		waddstr(fStatusLineWindow, "low");
	}

	wprintw(fStatusLineWindow, "  trace level: %d", fTraceLevel);
	wclrtoeol(fStatusLineWindow);
}

void CursesFrontend::DisplayDriveStatus(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			DisplayDriveNumber(d);
			DisplayDriveChangedStatus(d);
			DisplayDriveProtectedStatus(d);
			DisplayDriveDensity(d);
			DisplayDriveSectors(d);
			DisplayDriveFilename(d);
		}
	}
}

void CursesFrontend::DisplayDriveNumber(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			wmove(fDriveStatusWindow, fDriveYStart + d - DiskManager::eMinDriveNumber, fDriveNoX);
			wprintw(fDriveStatusWindow,"D%d:", d);
		}
	}
}

void CursesFrontend::DisplayDriveChangedStatus(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			wmove(fDriveStatusWindow, fDriveYStart + d - DiskManager::eMinDriveNumber, fDriveChangedX);
			if (fDiskManager->DriveInUse(d)) {
				if (fDiskManager->DriveIsChanged(d)) {
					wattrset(fDriveStatusWindow, fDriveColorChanged);
					waddch(fDriveStatusWindow, 'C');
					wattrset(fDriveStatusWindow, fDriveColorStandard);
				} else {
					waddch(fDriveStatusWindow, ' ');
				}
			} else {
				waddch(fDriveStatusWindow, '-');
			}
		}
	}
}

void CursesFrontend::DisplayDriveProtectedStatus(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			wmove(fDriveStatusWindow, fDriveYStart + d - DiskManager::eMinDriveNumber, fDriveProtectedX);
			if (fDiskManager->DriveInUse(d)) {
				if (fDiskManager->DriveIsWriteProtected(d)) {
					wattrset(fDriveStatusWindow, fDriveColorWriteProtected);
					waddch(fDriveStatusWindow, 'P');
					wattrset(fDriveStatusWindow, fDriveColorStandard);
				} else {
					wattrset(fDriveStatusWindow, fDriveColorWritable);
					waddch(fDriveStatusWindow, 'W');
					wattrset(fDriveStatusWindow, fDriveColorStandard);
				}
			} else {
				waddch(fDriveStatusWindow, '-');
			}
		}
	}
}

void CursesFrontend::DisplayDriveDensity(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			wmove(fDriveStatusWindow, fDriveYStart + d - DiskManager::eMinDriveNumber, fDriveDensityX);
			if (fDiskManager->DriveInUse(d)) {
				switch (fDiskManager->GetConstDiskImage(d)->GetSectorLength()) {
				case e128BytesPerSector:
					waddch(fDriveStatusWindow, 'S');
					break;
				case e256BytesPerSector:
					waddch(fDriveStatusWindow, 'D');
					break;
				default:
					waddch(fDriveStatusWindow, '?');
					break;
				}
			} else {
				waddch(fDriveStatusWindow, '-');
			}
		}
	}
}

void CursesFrontend::DisplayDriveSectors(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			wmove(fDriveStatusWindow, fDriveYStart + d - DiskManager::eMinDriveNumber, fDriveSectorsX);
			if (fDiskManager->DriveInUse(drive)) {
				wprintw(fDriveStatusWindow, "%5d", fDiskManager->GetConstDiskImage(d)->GetNumberOfSectors());
			} else {
				waddstr(fDriveStatusWindow, "-----");
			}
		}
	}
}

void CursesFrontend::DisplayDriveFilename(DiskManager::EDriveNumber drive)
{
	if ( (drive == DiskManager::eAllDrives) || fDiskManager->DriveNumberOK(drive) ) {
		int min=drive, max=drive;
		if (drive == DiskManager::eAllDrives) {
			min = DiskManager::eMinDriveNumber;
			max = DiskManager::eMaxDriveNumber;
		}
	
		for (int i = min; i<=max; i++) {
			DiskManager::EDriveNumber d = DiskManager::EDriveNumber(i);
			unsigned int len=0;
			const char* filename;

			if (fDiskManager->DriveInUse(d)) {
				filename = fDiskManager->GetImageFilename(d);
			} else {
				filename = "<empty>";
			}

			wmove(fDriveStatusWindow, fDriveYStart + d - DiskManager::eMinDriveNumber, fDriveFilenameX);
	
			if (filename) {
				char *sf = ShortenFilename(filename, fDriveFilenameLength);
				if (sf) {
					len = strlen(sf);
					waddstr(fDriveStatusWindow, sf);
					delete[] sf;
				} else {
					len = 0;
				}
			}
			if (len < fDriveFilenameLength) {
				wclrtoeol(fDriveStatusWindow);
			}
		}
	}
}

bool CursesFrontend::ShowCursor(bool on)
{
	bool oldStat = fCursorStatus;
	if (on) {
		if (curs_set(1) == ERR) {
			DPRINTF("enabeling cursor failed");
		}
	} else {
		if (curs_set(0) == ERR) {
			DPRINTF("disabeling cursor failed");
		}
	}
	fCursorStatus = on;
	return oldStat;
}

void CursesFrontend::AddLogLine(const char* str)
{
	StartLogLine();
	AddLogString(str);
	EndLogLine();
}

void CursesFrontend::StartLogLine()
{
	if (!fFirstLogLine) {
		wprintw(fLogWindow,"\n");
	}
	fFirstLogLine = false;
}

void CursesFrontend::EndLogLine()
{
}

void CursesFrontend::AddLogString(const char* str)
{
	wattrset(fLogWindow, fLogColorStandard);
	wprintw(fLogWindow, "%s", str);
}

void CursesFrontend::AddLogHighlightString(const char* str)
{
	wattrset(fLogWindow, fLogColorHighlight);
	wprintw(fLogWindow, "%s", str);
	wattrset(fLogWindow, fLogColorStandard);
}

void CursesFrontend::AddLogOKString(const char* str)
{
	wattrset(fLogWindow, fLogColorOK);
	wprintw(fLogWindow, "%s", str);
	wattrset(fLogWindow, fLogColorStandard);
}

void CursesFrontend::AddLogDebugString(const char* str)
{
	wattrset(fLogWindow, fLogColorDebug);
	wprintw(fLogWindow, "%s", str);
	wattrset(fLogWindow, fLogColorStandard);
}

void CursesFrontend::AddLogWarningString(const char* str)
{
	wattrset(fLogWindow, fLogColorWarning);
	wprintw(fLogWindow, "%s", str);
	wattrset(fLogWindow, fLogColorStandard);
}

void CursesFrontend::AddLogErrorString(const char* str)
{
	wattrset(fLogWindow, fLogColorError);
	wprintw(fLogWindow, "%s", str);
	wattrset(fLogWindow, fLogColorStandard);
}

void CursesFrontend::PrintLogLine(const char* __restrict format, ...)
{
	va_list arg;
	va_start(arg, format);

	if (!fFirstLogLine) {
		waddstr(fLogWindow, "\n");
	}
	vw_printw(fLogWindow, format, arg);
	va_end(arg);

	fFirstLogLine = false;
}

int CursesFrontend::GetCh(bool ignoreResize)
{
	int ret;
	int ch;
	do {
		ret = fDiskManager->DoServing(STDIN_FILENO);

		if (ret == 1 && GotSigWinCh() && !ignoreResize ) {
			ch = KEY_RESIZE;
			break;
		} else if (ret == 0) {
			ch = wgetch(fInputLineWindow);
			if (ch == KEY_RESIZE) {
				DPRINTF("got KEY_RESIZE from ncurses!");
			} else {
				break;
			}
		}
	} while (1);
	return ch;
}

bool CursesFrontend::IsAbortChar(int ch)
{
	return (ch == ERR) || (ch == 27) || (ch == 7)
		|| (ch == 'q') || (ch == 'Q');
}

DiskManager::EDriveNumber CursesFrontend::InputDriveNumber(bool allowAll)
{
	do {
		int ch = GetCh(true);
		if ( IsAbortChar(ch) ) {
			return DiskManager::eNoDrive;
		}
		if ( (ch >= '0' + DiskManager::eMinDriveNumber) &&
		     (ch <= '0' + DiskManager::eMaxDriveNumber) ) {
			return DiskManager::EDriveNumber(ch - '0');
		}
		if (allowAll && ( (ch == 'a') || (ch == 'A') ) ) {
			return DiskManager::eAllDrives;
		}
		beep();
	} while (1);
}

DiskManager::EDriveNumber CursesFrontend::InputUsedDriveNumber(bool allowAll)
{
	do {
		DiskManager::EDriveNumber d = InputDriveNumber(allowAll);
		if ( (d == DiskManager::eNoDrive) || (d == DiskManager::eAllDrives) ) {
			return d;
		}
		if ( fDiskManager->DriveInUse(d) ) {
			return d;
		}
		beep();
	} while (1);
}

DiskManager::EDriveNumber CursesFrontend::InputUnusedDriveNumber(bool allowAll)
{
	do {
		DiskManager::EDriveNumber d = InputDriveNumber(allowAll);
		if ( (d == DiskManager::eNoDrive) || (d == DiskManager::eAllDrives) ) {
			return d;
		}
		if ( ! fDiskManager->DriveInUse(d) ) {
			return d;
		}
		beep();
	} while (1);
}

CursesFrontend::EYesNo CursesFrontend::InputYesNo()
{
	int ch;
	do {
		ch = GetCh(true);
		if (IsAbortChar(ch)) {
			return eYN_Abort;
		}
		if ( (ch == 'y') || (ch == 'Y') ) {
			return eYN_Yes;
		}
		if ( (ch == 'n') || (ch == 'N') ) {
			return eYN_No;
		}
		beep();
	} while(1);
}

void CursesFrontend::SetTraceLevel(int level)
{
	fTraceLevel = level;
        SIOTracer* tracer = SIOTracer::GetInstance();
	tracer->SetTraceGroup(SIOTracer::eTraceVerboseCommands, level == 1 );
	tracer->SetTraceGroup(SIOTracer::eTraceCommands, level >= 2 );
	tracer->SetTraceGroup(SIOTracer::eTraceUnhandeledCommands, level >= 2 );
	tracer->SetTraceGroup(SIOTracer::eTraceAtpInfo, level >= 2 );
	tracer->SetTraceGroup(SIOTracer::eTraceDataBlocks, level >= 3 );
}

void CursesFrontend::ShowStandardHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"press 'h' for help, 'q' to quit");
}

void CursesFrontend::ShowYesNoHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow, "'y'=yes, 'n'=no, '^G','q'=abort");
}

void CursesFrontend::ShowPagerHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"press <space> to continue, '^G','q' to abort");
}

void CursesFrontend::ShowDriveInputHint(EDriveInputHintType type)
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	wprintw(fBottomLineWindow,"drive '%d'..'%d'",
		DiskManager::eMinDriveNumber, DiskManager::eMaxDriveNumber);
	switch (type) {
	case eDriveInputAll:
		waddstr(fBottomLineWindow,", 'a'=all drives");
		break;
	case eDriveInputAllChanged:
		waddstr(fBottomLineWindow,", 'a'=all changed drives");
		break;
	default:
		break;
	}
	waddstr(fBottomLineWindow,", '^G','q'=abort");
}

void CursesFrontend::ShowHint(const char* string)
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	wprintw(fBottomLineWindow,"%s", string);
}

void CursesFrontend::ShowFileInputHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	wprintw(fBottomLineWindow,"enter filename  <tab>=complete, '^F'=file-select, '^G'=abort");
}

void CursesFrontend::ShowTraceLevelHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"level '0'..'3', '^G','q'=abort");
}

void CursesFrontend::ShowCreateDriveHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"'1'=90k '2'=130k '3'=180k '4'=360k 's'=custom SD 'd'=custom DD '^G','q'=abort");
}

void CursesFrontend::ShowImageSizeHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"'1'..'65535' sectors, '^G'=abort");
}

void CursesFrontend::ShowHighSpeedHint()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"'0'=slow '1'=high '2'=high with pauses '^G','q'=abort");
}

void CursesFrontend::AbortInput()
{
	ShowCursor(false);
	ShowStandardHint();
	InitTopLine();
	werase(fInputLineWindow);
	UpdateScreen();
}

void CursesFrontend::ClearInputLine()
{
	werase(fInputLineWindow);
	wmove(fInputLineWindow, 0, 0);
}

void CursesFrontend::ShowDriveNumber(DiskManager::EDriveNumber d, bool allChanged)
{
	switch (d) {
	case DiskManager::eNoDrive:
		waddstr(fInputLineWindow, "<aborted>");
		break;
	case DiskManager::eAllDrives:
		waddstr(fInputLineWindow, "all");
		if (allChanged) {
			waddstr(fInputLineWindow, " changed");
		}
		break;
	default:
		wprintw(fInputLineWindow, "%d", d);
		break;
	}
}

void CursesFrontend::ShowYesNo(EYesNo yn)
{
	switch (yn) {
	case eYN_Abort:
		waddstr(fInputLineWindow, "<aborted>");
		break;
	case eYN_Yes:
		waddstr(fInputLineWindow, "yes");
		break;
	case eYN_No:
		waddstr(fInputLineWindow, "no");
		break;
	}
}

void CursesFrontend::InternalProcessWriteProtectDrive(bool protect)
{
	ShowDriveInputHint(eDriveInputAll);
	ClearInputLine();
	if (protect) {
		waddstr(fInputLineWindow, "write protect disk: ");
	} else {
		waddstr(fInputLineWindow, "un-protect disk: ");
	}
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d = InputUsedDriveNumber(true);

	ShowCursor(false);

	if (d == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	ShowDriveNumber(d);

	fDiskManager->SetWriteProtectImage(d, protect);
	DisplayDriveProtectedStatus(d);
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessWriteProtectDrive()
{
	InternalProcessWriteProtectDrive(true);
}

void CursesFrontend::ProcessUnprotectDrive()
{
	InternalProcessWriteProtectDrive(false);
}

bool CursesFrontend::ProcessQuit()
{
	bool ret = true;
	if (fDiskManager->CheckForChangedImages()) {
		ShowYesNoHint();
		ClearInputLine();
		waddstr(fInputLineWindow, "there are changed images - really quit? ");
		ShowCursor(true);
		UpdateScreen();

		EYesNo yn = InputYesNo();

		switch (yn) {
		case eYN_Abort:
			AbortInput();
			return false;
		case eYN_Yes:
			ShowYesNo(yn);
			ret = true;
			break;
		case eYN_No:
			ShowYesNo(yn);
			ret = false;
			break;
		}
	}

	ShowCursor(false);
	DisplayStatusLine();
	ShowStandardHint();
	UpdateScreen();

	return ret;
}

void CursesFrontend::ProcessExchangeDrives()
{
	ShowDriveInputHint(eDriveInputStandard);
	ClearInputLine();
	waddstr(fInputLineWindow, "exchange drive: ");
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d1 = InputDriveNumber(true);

	if (d1 == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	wprintw(fInputLineWindow, "%d with drive: ", d1);

	DiskManager::EDriveNumber d2;
	do {
		d2 = InputDriveNumber(true);
		if (d2 == d1) {
			beep();
		}
	} while (d1 == d2);

	if (d2 == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	wprintw(fInputLineWindow, "%d", d2);

	ShowCursor(false);
	fDiskManager->ExchangeDrives(d1, d2);
	DisplayDriveStatus(d1);
	DisplayDriveStatus(d2);
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessUnloadDrive()
{
	ShowDriveInputHint(eDriveInputAll);
	ClearInputLine();
	waddstr(fInputLineWindow, "unload drive: ");
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d = InputUsedDriveNumber(true);

	if (d == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	ShowDriveNumber(d);

	bool doUnload = true;

	if (fDiskManager->DriveIsChanged(d)) {
		ClearInputLine();
		if ( d == DiskManager::eAllDrives ) {
			waddstr(fInputLineWindow,"there are changed images - really unload them? ");
		} else {
			wprintw(fInputLineWindow, "drive %d is changed - really unload it? ", d);
		}
		ShowYesNoHint();
		ShowCursor(true);
		UpdateScreen();

		EYesNo yn = InputYesNo();

		switch (yn) {
		case eYN_Abort:
			AbortInput();
			return;
		case eYN_Yes:
			ShowYesNo(yn);
			break;
		case eYN_No:
			ShowYesNo(yn);
			doUnload = false;
			break;
		}
	}

	if (doUnload) {
		if (!fDiskManager->UnloadDiskImage(d)) {
			DPRINTF("unloading disk image %d failed!", d);
		}
	}

	ShowCursor(false);
	DisplayDriveStatus(d);
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessLoadDrive()
{
	ShowDriveInputHint(eDriveInputStandard);
	ClearInputLine();
	waddstr(fInputLineWindow, "load drive: ");
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d = InputUnusedDriveNumber(false);

	if (d == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	ShowDriveNumber(d);

	waddstr(fInputLineWindow," file: ");
	unsigned int x,y, width;
	getyx(fInputLineWindow, y, x);
	if ( (x + 10) > fScreenWidth ) {
		x = 0;
		width = fScreenWidth;
	} else {
		width = fScreenWidth - x;
	}

	FileInput input(this, fDirCache);
	fDirCache->ClearCache();

	char filename[PATH_MAX+1];
	filename[0] = 0;

	ShowFileInputHint();

	StringInput::EInputResult res = 
	input.InputString(fInputLineWindow, x, y, width, filename, PATH_MAX+1, eFileSelectKey, true, &fFilenameHistory);

	switch (res) {
	case StringInput::eInputStartSelect:
		{
			// DPRINTF("start select: path=\"%s\" filename=\"%s\"", input.GetPath(), input.GetFilename());
			ClearInputLine();

			FileSelect sel(this);
			ShowAuxWindow(true);
			bool ok=sel.Select(fAuxWindow, input.GetPath(), input.GetFilename(), filename, PATH_MAX+1);
			ShowAuxWindow(false);
			if (!ok) {
				AbortInput();
				return;
			}
			ClearInputLine();
		}
		break;
	case StringInput::eInputAborted:
		AbortInput();
		return;
	default:
		break;
	}

	if (filename[0] == 0) {
		AbortInput();
		return;
	}
	if (fDiskManager->LoadDiskImage(d, filename)) {
		const char* fn = fDiskManager->GetImageFilename(d);
		if (fn == NULL) {
			DPRINTF("getting filename of newly loaded image failed!");
		} else {
			fFilenameHistory.Add(fn);
		}
	}

	ShowCursor(false);
	DisplayDriveStatus(d);
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessWriteDrive()
{
	ShowDriveInputHint(eDriveInputAllChanged);
	ClearInputLine();
	waddstr(fInputLineWindow, "write drive: ");
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d = InputUsedDriveNumber(true);

	if (d == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	ShowDriveNumber(d, true);

	if (d == DiskManager::eAllDrives) {
		fDiskManager->WriteBackImagesIfChanged();
	} else {
		waddstr(fInputLineWindow," file: ");
		unsigned int x,y, width;
		getyx(fInputLineWindow, y, x);
		if ( (x + 10) > fScreenWidth ) {
			x = 0;
			width = fScreenWidth;
		} else {
			width = fScreenWidth - x;
		}

		FileInput input(this, fDirCache);
		fDirCache->ClearCache();

		char filename[PATH_MAX+1];
		const char* imgfn = fDiskManager->GetImageFilename(d);
		if (imgfn) {
			strncpy(filename, imgfn, PATH_MAX+1);
			filename[PATH_MAX+1] = 0;
		} else {
			filename[0] = 0;
		}

		ShowFileInputHint();

		StringInput::EInputResult res = 
		input.InputString(fInputLineWindow, x, y, width, filename, PATH_MAX+1, eFileSelectKey, true, &fFilenameHistory);

		switch (res) {
		case StringInput::eInputStartSelect:
			{
				// DPRINTF("start select: path=\"%s\" filename=\"%s\"", input.GetPath(), input.GetFilename());
				ClearInputLine();

				FileSelect sel(this);
				ShowAuxWindow(true);
				bool ok=sel.Select(fAuxWindow, input.GetPath(), input.GetFilename(), filename, PATH_MAX+1);
				ShowAuxWindow(false);
				if (!ok) {
					AbortInput();
					return;
				}
			}
			break;
		case StringInput::eInputAborted:
			AbortInput();
			return;
		default:
			break;
		}

		if (filename[0] == 0) {
			AbortInput();
			return;
		}
		if (fDiskManager->WriteDiskImage(d, filename)) {
			const char* fn = fDiskManager->GetImageFilename(d);
			if (fn == NULL) {
				DPRINTF("getting filename of newly loaded image failed!");
			} else {
				fFilenameHistory.Add(fn);
			}
		}
	}

	ShowCursor(false);
	DisplayDriveStatus(d);
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessCreateDrive()
{
	ShowDriveInputHint(eDriveInputStandard);
	ClearInputLine();
	waddstr(fInputLineWindow, "create drive: ");
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d = InputUnusedDriveNumber(false);

	if (d == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	ShowDriveNumber(d);

	ShowCreateDriveHint();
	waddstr(fInputLineWindow," size: ");
	UpdateScreen();

	int ch;
	do {
		ch = GetCh(true);
	} while ( (!IsAbortChar(ch)) &&
		  (ch != '1') && (ch != '2') && (ch != '3') && (ch != '4') &&
		  (ch != 's') && (ch != 'S') && (ch != 'd') && (ch != 'D') );

	if (IsAbortChar(ch)) {
		AbortInput();
		return;
	}
	switch (ch) {
	case '1':
		waddstr(fInputLineWindow,"90k");
		fDiskManager->CreateAtrMemoryImage(d, e90kDisk);
		break;
	case '2':
		waddstr(fInputLineWindow,"130k");
		fDiskManager->CreateAtrMemoryImage(d, e130kDisk);
		break;
	case '3':
		waddstr(fInputLineWindow,"180k");
		fDiskManager->CreateAtrMemoryImage(d, e180kDisk);
		break;
	case '4':
		waddstr(fInputLineWindow,"360k");
		fDiskManager->CreateAtrMemoryImage(d, e360kDisk);
		break;
	case 's':
	case 'S':
	case 'd':
	case 'D':
		{
			bool isDD = false;
			waddstr(fInputLineWindow,"custom ");
			if ((ch == 's') || (ch == 'S')) {
				waddstr(fInputLineWindow,"SD");
			} else {
				waddstr(fInputLineWindow,"DD");
				isDD = true;
			}

			waddstr(fInputLineWindow," sectors: ");

			unsigned int x,y, width;
			getyx(fInputLineWindow, y, x);
			width = 6;
			if ( (x + 6) > fScreenWidth ) {
				x = 0;
				if (fScreenWidth < 6) {
					width = fScreenWidth;
				}
			}
	
			StringInput input(this);

			char sectors[6];
			sectors[0] = 0;
			int numSectors;

			ShowImageSizeHint();

			do {
				StringInput::EInputResult res = 
					input.InputString(fInputLineWindow, x, y, width, sectors, 6, -1, true, &fImageSizeHistory);
				if (res == StringInput::eInputAborted) {
					AbortInput();
					return;
				}
				numSectors = atoi(sectors);
				if ( (numSectors >= 1) && (numSectors <= 65535) ) {
					break;
				} else {
					beep();
				}
			} while (1);

			snprintf(sectors, 6, "%d", numSectors);
			fImageSizeHistory.Add(sectors);

			if (isDD) {
				fDiskManager->CreateAtrMemoryImage(d, e256BytesPerSector, numSectors);
			} else {
				fDiskManager->CreateAtrMemoryImage(d, e128BytesPerSector, numSectors);
			}
		}
		break;
	default:
		DPRINTF("unhandeled option %d in ProcessCreateImage", ch);
	}

	ShowCursor(false);
	DisplayDriveStatus(d);
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessSetTraceLevel()
{
	ShowTraceLevelHint();
	ClearInputLine();
	waddstr(fInputLineWindow, "set trace level: ");
	ShowCursor(true);
	UpdateScreen();

	int ch;
	do {
		ch = GetCh(true);
		if (IsAbortChar(ch)) {
			AbortInput();
			return;
		}
		if ( (ch >= '0') && (ch <= '3') ) {
			break;
		}
		beep();
	} while(1);

	int level = ch - '0';

	wprintw(fInputLineWindow, "%d", level);

	ShowCursor(false);
	SetTraceLevel(level);
	DisplayStatusLine();
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessSetCableType()
{
	werase(fBottomLineWindow);
	wmove(fBottomLineWindow, 0, 0);
	waddstr(fBottomLineWindow,"'r'=RING, 'd'=DSR, 'c'=CTS, 'q'=abort");
	ClearInputLine();
	waddstr(fInputLineWindow, "set cable type: ");
	ShowCursor(true);
	UpdateScreen();

	int ch;
	do {
		ch = GetCh(true);
		if (IsAbortChar(ch)) {
			AbortInput();
			return;
		}
		if ( (ch == 'R') || (ch == 'r') ||
		     (ch == 'D') || (ch == 'd') ||
		     (ch == 'C') || (ch == 'c') ) {
			break;
		}
		beep();
	} while(1);

	switch (ch) {
	case 'R':
	case 'r':
		wprintw(fInputLineWindow, "RING");
		fDiskManager->SetSioServerMode(SIOWrapper::eCommandLine_RI);
		break;
	case 'D':
	case 'd':
		wprintw(fInputLineWindow, "DSR");
		fDiskManager->SetSioServerMode(SIOWrapper::eCommandLine_DSR);
		break;
	case 'C':
	case 'c':
		wprintw(fInputLineWindow, "CTS");
		fDiskManager->SetSioServerMode(SIOWrapper::eCommandLine_CTS);
		break;
	}

	ShowCursor(false);
	DisplayStatusLine();
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessSetHighSpeed()
{
	ShowHighSpeedHint();
	ClearInputLine();
	waddstr(fInputLineWindow, "set SIO speed: ");
	ShowCursor(true);
	UpdateScreen();

	int ch;
	do {
		ch = GetCh(true);
		if (IsAbortChar(ch)) {
			AbortInput();
			return;
		}
		if ( (ch == '0') || (ch == '1') || (ch == '2') ) {
			break;
		}
		beep();
	} while(1);

	switch (ch) {
	case '0':
		waddstr(fInputLineWindow,"slow");
		fDiskManager->SetHighSpeedMode(DiskManager::eHighSpeedOff);
		break;
	case '1':
		waddstr(fInputLineWindow,"high");
		fDiskManager->SetHighSpeedMode(DiskManager::eHighSpeedOn);
		break;
	case '2':
		waddstr(fInputLineWindow,"high with pause");
		fDiskManager->SetHighSpeedMode(DiskManager::eHighSpeedWithPause);
		break;
	}

	ShowCursor(false);
	DisplayStatusLine();
	ShowStandardHint();
	UpdateScreen();
}

void CursesFrontend::ProcessShowDirectory()
{
	ShowDriveInputHint(eDriveInputAll);
	ClearInputLine();
	waddstr(fInputLineWindow, "directory of drive: ");
	ShowCursor(true);
	UpdateScreen();

	DiskManager::EDriveNumber d = InputUsedDriveNumber(false);

	ShowCursor(false);

	if (d == DiskManager::eNoDrive) {
		AbortInput();
		return;
	}

	ShowDriveNumber(d);

	ShowDos2Directory(fDiskManager->GetDiskImage(d));

	ShowStandardHint();
	InitTopLine();

	UpdateScreen();
}

static const char* const helpText[] =
	{
		"l     load image into drive",
		"w     write drive(s) to image(s)",
		"c     create drive",
		"u     unload drive",
		"x     exchange drives",
		"p     write protect drive(s)",
		"P     write un-protect drive(s)",
		"",
		"d     display DOS 2.x directory of drive",
		"",
	        "t     set trace level",
		"C     set cable type",
		"s     set SIO high speed mode",
		"",
		"^L    redraw screen",
		"h     show help screen",
		"q     quit atariserver",
		0
	};

void CursesFrontend::ProcessShowHelp()
{
	ClearInputLine();
	ShowText("[ help screen ]", helpText);
	ShowStandardHint();
	InitTopLine();
	UpdateScreen();
}

void CursesFrontend::AddFilenameHistory(const char* string)
{
	fFilenameHistory.Add(string);
}

bool CursesFrontend::ShowDos2Directory(const RCPtr<DiskImage>& image, bool beQuiet)
{
	if (image.IsNull()) {
		return false;
	}

	RCPtr<Dos2Dir> dir = image->GetDos2Directory(beQuiet);
	if (dir.IsNull()) {
		return false;
	}

	bool oldAux = ShowAuxWindow(true);
	bool oldCursor = ShowCursor(false);

	unsigned int columns, lines;
	getmaxyx(fAuxWindow, lines, columns);

	unsigned int cols = columns / Dos2Dir::eFileWidth;
	if (cols < 1) {
		cols = 1;
	}

	SetTopLineFilename(image->GetFilename(), false);

	ShowPagerHint();

	unsigned int x,y;
	unsigned int filenum = 0;

	bool done = false;

	while (!done) {
		unsigned int startFilenum = filenum;
		x = 0;
		y = 0;
		werase(fAuxWindow);

		while ( (filenum <= dir->GetNumberOfFiles() + 1) && (y < lines) ) {
			if (filenum == dir->GetNumberOfFiles()) {
				if (x != 0) {
					DPRINTF("DisplayDos2Directory: x != 0!");
				}
				wmove(fAuxWindow, y, 0);
				wprintw(fAuxWindow, "%d free sectors", dir->GetFreeSectors());
			} else {
				wmove(fAuxWindow, y, x*Dos2Dir::eFileWidth);
				waddstr(fAuxWindow, dir->GetFile(filenum));
			}

			filenum++;
			if (filenum == dir->GetNumberOfFiles()) {
				x=0;
				y++;
			} else if (filenum == dir->GetNumberOfFiles() + 1) {
				done = true;
			} else {
				x++;
				if (x==cols) {
					x = 0;
					y ++;
				}
			}
		}
		UpdateScreen();
		int ch = GetCh(false);
		if (ch == KEY_RESIZE) {
			HandleResize(true);
			ShowPagerHint();
			SetTopLineFilename(image->GetFilename(), false);
			ClearInputLine();
			done = false;
			filenum = startFilenum;

			getmaxyx(fAuxWindow, lines, columns);
			cols = columns / Dos2Dir::eFileWidth;
			if (cols < 1) {
				cols = 1;
			}
		} else {
			if (IsAbortChar(ch)) {
				done = true;
			}
		}
	}

	werase(fAuxWindow);
	ShowStandardHint();
	ShowAuxWindow(oldAux);
	ShowCursor(oldCursor);
	return true;
}

void CursesFrontend::ShowText(const char* title, const char* const * text)
{
	if (text == NULL) {
		return;
	}

	bool oldAux = ShowAuxWindow(true);
	bool oldCursor = ShowCursor(false);

	unsigned int columns, lines;
	getmaxyx(fAuxWindow, lines, columns);

	unsigned int currentLine;
	unsigned int currentTextLine = 0;
	const char* currentTextPtr = text[currentTextLine];
	unsigned int len;

	bool done = false;

	if (title) {
		SetTopLineString(title);
	}
	ShowPagerHint();

	while (!done) {
		currentLine = 0;
		unsigned int startTextLine = currentTextLine;
		const char* startTextPtr = currentTextPtr;
		werase(fAuxWindow);

		while ( currentTextPtr && (currentLine < lines) ) {
			len = strlen(currentTextPtr);
			if (len > columns) {
				len = columns;
				while ( (len>0) && (currentTextPtr[len-1] != ' ') && (currentTextPtr[len-1] != '-') ) {
					len--;
				}
				if (len == 0) {
					len = columns;
				}
			}
			wmove(fAuxWindow, currentLine, 0);
			if (len > 0) {
				waddnstr(fAuxWindow, currentTextPtr, len);
				currentTextPtr += len;
			}
			currentLine++;
			if ( *currentTextPtr == 0) {
				currentTextPtr = text[++currentTextLine];
				if (currentTextPtr == NULL) {
					done = true;
				}
			}
		}

		UpdateScreen();
		int ch = GetCh(false);
		if (ch == KEY_RESIZE) {
			HandleResize(true);
			ShowPagerHint();
			if (title) {
				SetTopLineString(title);
			}
			ClearInputLine();
			currentTextLine = startTextLine;
			currentTextPtr = startTextPtr;
			done = false;

			getmaxyx(fAuxWindow, lines, columns);
		} else {
			if (IsAbortChar(ch)) {
				done = true;
			}
			while (currentTextPtr && (*currentTextPtr == 0) ) {
				currentTextPtr = text[++currentTextLine];
			}
			if (currentTextPtr == NULL) {
				done = true;
			}
		}
	}

	werase(fAuxWindow);
	ShowAuxWindow(oldAux);
	ShowCursor(oldCursor);
}
