/*
   atariserver.cpp - implementation of an Atari SIO server, using
   a curses frontend

   Copyright (C) 2003 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "CursesFrontend.h"
#include "CursesFrontendTracer.h"
#include "SIOTracer.h"
#include "FileTracer.h"
#include "History.h"
#include "Error.h"
#include "AtariDebug.h"

#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sched.h>
#include <sys/mman.h>

static CursesFrontend* frontend= NULL;

static uid_t euid, uid;
static gid_t egid, gid;

static SIOTracer* sioTracer = 0;

static void my_sig_handler(int sig)
{
	switch (sig) {
	case SIGWINCH:
		if (frontend) {
			frontend->IndicateGotSigWinCh();
			//ungetch(KEY_RESIZE);
		}
		break;
	case SIGALRM:
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		if (sioTracer) {
			sioTracer->RemoveAllTracers();
		}
		endwin();
		exit(0);
	}
}

static void reserve_stack_memory()
{
#define RESERVE_STACK_SIZE 100000
	char dummy_string[RESERVE_STACK_SIZE];
	int i;
	for (i=0;i<RESERVE_STACK_SIZE;i++) {
		dummy_string[i]=0;
	}
}

static void set_realtime_scheduling(int priority)
{
	struct sched_param sp;
	pid_t myPid;

	if (seteuid(euid) != 0) {
		printf("cannot set euid back to %d\n", euid);
		exit(1);
	}
	if (setegid(egid) != 0) {
		printf("cannot set egid back to %d\n", egid);
		exit(1);
	}

	memset(&sp, 0, sizeof(struct sched_param));
	sp.sched_priority = sched_get_priority_max(SCHED_RR) - priority;
	myPid = getpid();

	if (sched_setscheduler(myPid, SCHED_RR, &sp) == 0) {
		LOG("activated realtime scheduling");
/*
#ifdef ATARISIO_DEBUG
	LOG("the server will automatically terminate in 5 minutes!");
	signal(SIGALRM, my_sig_handler);
	alarm(300);
#endif
*/
	} else {
		WARN("Cannot set realtime scheduling! please run as root!");
	}

	reserve_stack_memory();
	if (mlockall(MCL_CURRENT) == 0) {
		LOG("mlockall(2) succeeded");
	} else {
		WARN("mlockall(2) failed!");
	}

	// drop root privileges - set eid to real id
	if (seteuid(uid) != 0) {
		printf("cannot set euid to %d\n", uid);
		exit(1);
	}
	if (setegid(gid) != 0) {
		printf("cannot set egid to %d\n", gid);
		exit(1);
	}
}

static int trace_level = 0;

static void process_args(RCPtr<DiskManager>& manager, CursesFrontend* frontend, int argc, char** argv)
{
	bool write_protect_next = false;

	int drive = 1;

	for (int i=1;i<argc;i++) {
		if (argv[i]) {
			int len=strlen(argv[i]);
			if (len == 0) {
				endwin();
				printf("illegal argv!\n");
				exit(1);
			}
			if (argv[i][0]=='-') {
				switch(argv[i][1]) {
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
					if (len != 2) {
						goto illegal_option;
					}
					drive = argv[i][1] - '0';
					break;
				case 't':
					if (len != 2) {
						goto illegal_option;
					}
					trace_level++;
					break;
				case 's':
					manager->SetHighSpeedMode(DiskManager::eHighSpeedOff);
					LOG("disabling high-speed SIO");
					break;
				case 'S':
					manager->SetHighSpeedMode(DiskManager::eHighSpeedWithPause);
					LOG("enabling high-speed SIO mode with pauses");
					break;
				case 'c':
					manager->SetSioServerMode(SIOWrapper::eCommandLine_DSR);
					LOG("using alternative SIO2PC cable type (command=DSR)");
					break;
				case 'C':
					manager->SetSioServerMode(SIOWrapper::eCommandLine_CTS);
					LOG("using alternative SIO2PC/nullmodem cable type (command=CTS)");
					break;
				case 'p':
					write_protect_next = true;
					break;
				default:
illegal_option:
					ERROR("illegal option \"%s\"", argv[i]);
				}
			} else {
				if (manager->DriveNumberOK(DiskManager::EDriveNumber(drive))) {
					DiskManager::EDriveNumber driveNo = DiskManager::EDriveNumber(drive);
					if (manager->DriveInUse(driveNo)) {
						ERROR("drive D%d: already assigned!", drive);
					}
					if (!manager->LoadDiskImage(driveNo, argv[i])) {
						ERROR("cannot load \"%s\" into D%d:", argv[i], drive);
					} else {
						if (write_protect_next) {
							if (manager->SetWriteProtectImage(driveNo, true)) {
								ERROR("warning: write protecting D%d: failed!", drive);
							}
							write_protect_next = false;
						}
	
						frontend->AddFilenameHistory(manager->GetImageFilename(driveNo));
						drive++;
					}
				} else {
					ERROR("too many images - there is no drive D%d:",drive);
				}
			}
		}
	}
}

static void SetDefaultTraceLevels(const RCPtr<AbstractTracer>& tracer)
{
	sioTracer->SetTraceGroup(SIOTracer::eTraceInfo, true, tracer);
	sioTracer->SetTraceGroup(SIOTracer::eTraceDebug, true, tracer);
}

int main(int argc, char** argv)
{
	RCPtr<DiskManager> manager;
	try {
		manager = new DiskManager;
	}
	catch (ErrorObject& err) {
		fprintf(stderr,"%s\n", err.AsString());
		//fprintf(stderr,"euid = %d egid = %d\n", geteuid(), getegid());
		exit(1);
	}

	euid = geteuid();
	uid = getuid();
	egid = getegid();
	gid = getgid();

	// drop root privileges
	if (seteuid(uid) != 0) {
		printf("cannot set euid to %d\n", uid);
		exit(1);
	}
	if (setegid(gid) != 0) {
		printf("cannot set egid to %d", gid);
		exit(1);
	}

	bool wantHelp = false;
	bool useColor = true;
	const char* traceFile = 0;

	// scan argv for "-h", "-m", -"o file"
	{
		for (int i=1; i<argc; i++) {
			if ( (argv[i][0] == '-') && (argv[i][2] == 0) ) {
				switch (argv[i][1]) {
				case 'h':
					wantHelp = true;
					argv[i] = 0;
					break;
				case 'm':
					useColor = false;
					argv[i] = 0;
					break;
				case 'o':
					if (i+1 < argc) {
						if (traceFile) {
							printf("only one trace file allowed - ignoring -f %s\n",
								argv[i+1]);
						} else {
							traceFile = argv[i+1];
						}
						argv[i] = 0;
						argv[i+1] = 0;
						i++;
					}
					break;
				default:
					break;
				}
			}
		}
	}

	if (wantHelp) {
		printf("usage: [-cChmsSt] [-o file] [ [-1] [-p] file1 [ [-2] [-p] file2 ...] ]\n");
		printf("-h          display help\n");
		printf("-c          use alternative SIO2PC cable (command=DSR)\n");
		printf("-C          use alternative SIO2PC/nullmodem cable (command=CTS)\n");
		printf("-m          monochrome mode\n");
		printf("-o file     save trace output to <file>\n");
		printf("-p          write protect the next image\n");
		printf("-s          slow mode - disable highspeed SIO\n");
		printf("-S          high speed SIO mode with pauses\n");
		printf("-t          increase SIO trace level (default:0, max:3)\n");
		printf("-1..-8      set current drive number (default: 1)\n"); 
		printf("<filename>  load <filename> into current drive number, and then\n");
		printf("            increment drive number by one\n");
		return 0;
	}

	signal(SIGWINCH, my_sig_handler);

	frontend = new CursesFrontend(manager, useColor);

	sioTracer = SIOTracer::GetInstance();
	{
		RCPtr<CursesFrontendTracer> cursesTracer(new CursesFrontendTracer(frontend));
		sioTracer->AddTracer(cursesTracer);
		SetDefaultTraceLevels(cursesTracer);
		sioTracer->SetTraceGroup(SIOTracer::eTraceImageStatus, true, cursesTracer);

		if (traceFile) {
			RCPtr<FileTracer> tracer;
			try {
				tracer = new FileTracer(traceFile);
				sioTracer->AddTracer(tracer);
				SetDefaultTraceLevels(tracer);
			}
			catch (ErrorObject& err) {
				ERROR("%s", err.AsString());
			}
		}
	}

	signal(SIGTERM, my_sig_handler);
	signal(SIGINT, my_sig_handler);
	signal(SIGHUP, my_sig_handler);

	set_realtime_scheduling(0);

	process_args(manager, frontend, argc, argv);
	frontend->DisplayDriveStatus();
	frontend->UpdateScreen();

	frontend->SetTraceLevel(trace_level);
	frontend->DisplayStatusLine();
	frontend->UpdateScreen();

	bool running = true;
	do {
		int ch = frontend->GetCh(false);
		switch (ch) {
		case 'h':
			frontend->ProcessShowHelp();
			break;
		case 'q':
			if (frontend->ProcessQuit()) {
				running = false;
			}
			break;
		case 'l':
			frontend->ProcessLoadDrive();
			break;
		case 'w':
			frontend->ProcessWriteDrive();
			break;
		case 'c':
			frontend->ProcessCreateDrive();
			break;
		case 'p':
			frontend->ProcessWriteProtectDrive();
			break;
		case 'P':
			frontend->ProcessUnprotectDrive();
			break;
		case 'x':
			frontend->ProcessExchangeDrives();
			break;
		case 'u':
			frontend->ProcessUnloadDrive();
			break;
		case 't':
			frontend->ProcessSetTraceLevel();
			break;
		case 'C':
			frontend->ProcessSetCableType();
			break;
		case 's':
			frontend->ProcessSetHighSpeed();
			break;
		case 'd':
			frontend->ProcessShowDirectory();
			break;
		case 12:
			endwin();
			frontend->InitWindows();
			break;
		case KEY_RESIZE:
			frontend->HandleResize();
			break;
		default:
			beep();
			//DPRINTF("got character 0x%04x ('%c')", ch, ( (ch>=32) && (ch<=126) ) ? ch : '.');
			break;
		}
		if (frontend->GotSigWinCh()) {
			frontend->HandleResize();
		}

	} while (running);

	sioTracer->RemoveAllTracers();
	{
		CursesFrontend* fe = frontend;
		frontend = NULL;
		delete fe;
	}
	return 0;
}
